@@ -1,6 +1,7 @@
 #include "common/common_pch.h"
 
 #include <QDebug>
+#include <QRegularExpression>
 #include <QScreen>
 #include <QSettings>
 #include <QSplitter>
@@ -11,34 +12,66 @@
 #include "common/qt.h"
 #include "common/version.h"
 #include "mkvtoolnix-gui/app.h"
+#include "mkvtoolnix-gui/jobs/program_runner.h"
+#include "mkvtoolnix-gui/merge/enums.h"
+#include "mkvtoolnix-gui/merge/source_file.h"
 #include "mkvtoolnix-gui/util/file_dialog.h"
 #include "mkvtoolnix-gui/util/settings.h"
+#include "mkvtoolnix-gui/util/string.h"
+#include "mkvtoolnix-gui/util/widget.h"
 
 namespace mtx { namespace gui { namespace Util {
 
-namespace {
-
-QFont
-defaultUiFont() {
-  auto font         = App::font();
-  // auto logicalDPI   = App::primaryScreen()->logicalDotsPerInch();
-  // auto minPointSize = static_cast<int>(9 + (std::max(logicalDPI, 95.) - 95 + 5) * 3 / 25);
-
-  // font.setPointSize(std::max(font.pointSize(), minPointSize));
-
-#if defined(SYS_WINDOWS)
-  font.setFamily(Q("Segoe UI"));
-#endif
-
-  return font;
-}
-
+QString
+Settings::RunProgramConfig::validate()
+  const {
+  return (m_type == RunProgramType::ExecuteProgram) && (m_commandLine.isEmpty() || m_commandLine.value(0).isEmpty()) ? QY("The program to execute hasn't been set yet.")
+       : (m_type == RunProgramType::PlayAudioFile)  && m_audioFile.isEmpty()                                         ? QY("The audio file to play hasn't been set yet.")
+       :                                                                                                               QString{};
 }
 
 bool
 Settings::RunProgramConfig::isValid()
   const {
-  return !m_commandLine.value(0).isEmpty();
+  return validate().isEmpty();
+}
+
+QString
+Settings::RunProgramConfig::name()
+  const {
+  if (!m_name.isEmpty())
+    return m_name;
+
+  return m_type == RunProgramType::ExecuteProgram    ? nameForExternalProgram()
+       : m_type == RunProgramType::PlayAudioFile     ? nameForPlayAudioFile()
+       : m_type == RunProgramType::ShutDownComputer  ? QY("Shut down the computer")
+       : m_type == RunProgramType::HibernateComputer ? QY("Hibernate the computer")
+       : m_type == RunProgramType::SleepComputer     ? QY("Sleep the computer")
+       :                                               Q("unknown");
+}
+
+QString
+Settings::RunProgramConfig::nameForExternalProgram()
+  const {
+  if (m_commandLine.isEmpty())
+    return QY("Execute a program");
+
+  auto program = m_commandLine.value(0);
+  program.replace(QRegularExpression{Q(".*[/\\\\]")}, Q(""));
+
+  return QY("Execute program '%1'").arg(program);
+}
+
+QString
+Settings::RunProgramConfig::nameForPlayAudioFile()
+  const {
+  if (m_audioFile.isEmpty())
+    return QY("Play an audio file");
+
+  auto audioFile = m_audioFile;
+  audioFile.replace(QRegularExpression{Q(".*[/\\\\]")}, Q(""));
+
+  return QY("Play audio file '%1'").arg(audioFile);
 }
 
 Settings Settings::s_settings;
@@ -69,6 +102,32 @@ Settings::iniFileLocation() {
 #endif
 }
 
+QString
+Settings::cacheDirLocation(QString const &subDir) {
+  QString dir;
+
+#if defined(SYS_WINDOWS)
+  if (!App::isInstalled())
+    dir = Q("%1/cache").arg(App::applicationDirPath());
+#endif
+
+  if (dir.isEmpty())
+    dir = QStandardPaths::writableLocation(QStandardPaths::CacheLocation);
+
+  if (!subDir.isEmpty())
+    dir = Q("%1/%2").arg(dir).arg(subDir);
+
+  return QDir::toNativeSeparators(dir);
+}
+
+QString
+Settings::prepareCacheDir(QString const &subDir) {
+  auto dirName = cacheDirLocation(subDir);
+  QDir{dirName}.mkpath(dirName);
+
+  return QDir::toNativeSeparators(dirName);
+}
+
 QString
 Settings::iniFileName() {
   return Q("%1/mkvtoolnix-gui.ini").arg(iniFileLocation());
@@ -160,6 +219,7 @@ Settings::load() {
   auto regPtr      = registry();
   auto &reg        = *regPtr;
   auto defaultFont = defaultUiFont();
+  boost::optional<QVariant> enableMuxingTracksByTheseTypes;
 
   reg.beginGroup("info");
   auto guiVersion                      = reg.value("guiVersion").toString();
@@ -167,6 +227,7 @@ Settings::load() {
 
   reg.beginGroup("settings");
   m_priority                           = static_cast<ProcessPriority>(reg.value("priority", static_cast<int>(NormalPriority)).toInt());
+  m_probeRangePercentage               = reg.value("probeRangePercentage", 0.3).toDouble();
   m_tabPosition                        = static_cast<QTabWidget::TabPosition>(reg.value("tabPosition", static_cast<int>(QTabWidget::North)).toInt());
   m_lastOpenDir                        = QDir{reg.value("lastOpenDir").toString()};
   m_lastOutputDir                      = QDir{reg.value("lastOutputDir").toString()};
@@ -176,23 +237,32 @@ Settings::load() {
   m_oftenUsedCountries                 = reg.value("oftenUsedCountries").toStringList();
   m_oftenUsedCharacterSets             = reg.value("oftenUsedCharacterSets").toStringList();
 
+  m_oftenUsedLanguagesOnly             = reg.value("oftenUsedLanguagesOnly",     false).toBool();;
+  m_oftenUsedCountriesOnly             = reg.value("oftenUsedCountriesOnly",     false).toBool();;
+  m_oftenUsedCharacterSetsOnly         = reg.value("oftenUsedCharacterSetsOnly", false).toBool();;
+
   m_scanForPlaylistsPolicy             = static_cast<ScanForPlaylistsPolicy>(reg.value("scanForPlaylistsPolicy", static_cast<int>(AskBeforeScanning)).toInt());
   m_minimumPlaylistDuration            = reg.value("minimumPlaylistDuration", 120).toUInt();
 
   m_setAudioDelayFromFileName          = reg.value("setAudioDelayFromFileName", true).toBool();
   m_autoSetFileTitle                   = reg.value("autoSetFileTitle",          true).toBool();
+  m_autoClearFileTitle                 = reg.value("autoClearFileTitle",        m_autoSetFileTitle).toBool();
   m_clearMergeSettings                 = static_cast<ClearMergeSettingsAction>(reg.value("clearMergeSettings", static_cast<int>(ClearMergeSettingsAction::None)).toInt());
   m_disableCompressionForAllTrackTypes = reg.value("disableCompressionForAllTrackTypes", false).toBool();
   m_disableDefaultTrackForSubtitles    = reg.value("disableDefaultTrackForSubtitles",    false).toBool();
+  m_mergeEnableDialogNormGainRemoval   = reg.value("mergeEnableDialogNormGainRemoval",   false).toBool();
   m_mergeAlwaysShowOutputFileControls  = reg.value("mergeAlwaysShowOutputFileControls",  true).toBool();
   m_mergeTrackPropertiesLayout         = static_cast<TrackPropertiesLayout>(reg.value("mergeTrackPropertiesLayout", static_cast<int>(TrackPropertiesLayout::HorizontalScrollArea)).toInt());
   m_mergeAddingAppendingFilesPolicy    = static_cast<MergeAddingAppendingFilesPolicy>(reg.value("mergeAddingAppendingFilesPolicy", static_cast<int>(MergeAddingAppendingFilesPolicy::Ask)).toInt());
+  m_mergeLastAddingAppendingDecision   = static_cast<MergeAddingAppendingFilesPolicy>(reg.value("mergeLastAddingAppendingDecision", static_cast<int>(MergeAddingAppendingFilesPolicy::Add)).toInt());
   m_headerEditorDroppedFilesPolicy     = static_cast<HeaderEditorDroppedFilesPolicy>(reg.value("headerEditorDroppedFilesPolicy", static_cast<int>(HeaderEditorDroppedFilesPolicy::Ask)).toInt());
 
-  m_uniqueOutputFileNames              = reg.value("uniqueOutputFileNames",     true).toBool();
   m_outputFileNamePolicy               = static_cast<OutputFileNamePolicy>(reg.value("outputFileNamePolicy", static_cast<int>(ToSameAsFirstInputFile)).toInt());
+  m_autoDestinationOnlyForVideoFiles   = reg.value("autoDestinationOnlyForVideoFiles", false).toBool();
   m_relativeOutputDir                  = QDir{reg.value("relativeOutputDir").toString()};
   m_fixedOutputDir                     = QDir{reg.value("fixedOutputDir").toString()};
+  m_uniqueOutputFileNames              = reg.value("uniqueOutputFileNames",   true).toBool();
+  m_autoClearOutputFileName            = reg.value("autoClearOutputFileName", m_outputFileNamePolicy != DontSetOutputFileName).toBool();
 
   m_enableMuxingTracksByLanguage       = reg.value("enableMuxingTracksByLanguage", false).toBool();
   m_enableMuxingAllVideoTracks         = reg.value("enableMuxingAllVideoTracks", true).toBool();
@@ -200,6 +270,9 @@ Settings::load() {
   m_enableMuxingAllSubtitleTracks      = reg.value("enableMuxingAllSubtitleTracks", false).toBool();
   m_enableMuxingTracksByTheseLanguages = reg.value("enableMuxingTracksByTheseLanguages").toStringList();
 
+  if (reg.contains("enableMuxingTracksByThesetypes"))
+    enableMuxingTracksByTheseTypes     = reg.value("enableMuxingTracksByThesetypes");
+
   m_useDefaultJobDescription           = reg.value("useDefaultJobDescription",       false).toBool();
   m_showOutputOfAllJobs                = reg.value("showOutputOfAllJobs",            true).toBool();
   m_switchToJobOutputAfterStarting     = reg.value("switchToJobOutputAfterStarting", false).toBool();
@@ -208,7 +281,6 @@ Settings::load() {
   m_removeOldJobs                      = reg.value("removeOldJobs",                                  true).toBool();
   m_removeOldJobsDays                  = reg.value("removeOldJobsDays",                              14).toInt();
 
-  m_disableAnimations                  = reg.value("disableAnimations", false).toBool();
   m_showToolSelector                   = reg.value("showToolSelector", true).toBool();
   m_warnBeforeClosingModifiedTabs      = reg.value("warnBeforeClosingModifiedTabs", true).toBool();
   m_warnBeforeAbortingJobs             = reg.value("warnBeforeAbortingJobs", true).toBool();
@@ -217,6 +289,7 @@ Settings::load() {
 
   m_chapterNameTemplate                = reg.value("chapterNameTemplate", QY("Chapter <NUM:2>")).toString();
   m_dropLastChapterFromBlurayPlaylist  = reg.value("dropLastChapterFromBlurayPlaylist", true).toBool();
+  m_ceTextFileCharacterSet             = reg.value("ceTextFileCharacterSet").toString();
 
   m_mediaInfoExe                       = reg.value("mediaInfoExe", Q("mediainfo-gui")).toString();
 
@@ -233,8 +306,48 @@ Settings::load() {
   reg.endGroup();               // settings
 
   loadDefaults(reg, guiVersion);
+  loadDerivingTrackLanguagesSettings(reg);
   loadSplitterSizes(reg);
+  loadDefaultInfoJobSettings(reg);
   loadRunProgramConfigurations(reg);
+  addDefaultRunProgramConfigurations(reg);
+  setDefaults(enableMuxingTracksByTheseTypes);
+}
+
+void
+Settings::setDefaults(boost::optional<QVariant> enableMuxingTracksByTheseTypes) {
+  if (m_oftenUsedLanguages.isEmpty())
+    for (auto const &languageCode : g_popular_language_codes)
+      m_oftenUsedLanguages << Q(languageCode);
+
+  if (m_oftenUsedCountries.isEmpty())
+    for (auto const &countryCode : g_popular_country_codes)
+      m_oftenUsedCountries << Util::mapToTopLevelCountryCode(Q(countryCode));
+
+  if (m_oftenUsedCharacterSets.isEmpty())
+    for (auto const &characterSet : g_popular_character_sets)
+      m_oftenUsedCharacterSets << Q(characterSet);
+
+  if (m_recognizedTrackLanguagesInFileNames.isEmpty())
+    for (auto const &languageCode : g_popular_language_codes)
+      m_recognizedTrackLanguagesInFileNames << Q(languageCode);
+
+  if (m_regexForDerivingTrackLanguagesFromFileNames.isEmpty())
+    m_regexForDerivingTrackLanguagesFromFileNames = mtx::gui::Merge::SourceFile::defaultRegexForDerivingLanguageFromFileName();
+
+  if (ToParentOfFirstInputFile == m_outputFileNamePolicy) {
+    m_outputFileNamePolicy = ToRelativeOfFirstInputFile;
+    m_relativeOutputDir    = Q("..");
+  }
+
+  m_enableMuxingTracksByTheseTypes.clear();
+  if (enableMuxingTracksByTheseTypes)
+    for (auto const &type : enableMuxingTracksByTheseTypes->toList())
+      m_enableMuxingTracksByTheseTypes << static_cast<Merge::TrackType>(type.toInt());
+
+  else
+    for (int type = static_cast<int>(Merge::TrackType::Min); type <= static_cast<int>(Merge::TrackType::Max); ++type)
+      m_enableMuxingTracksByTheseTypes << static_cast<Merge::TrackType>(type);
 }
 
 void
@@ -244,14 +357,30 @@ Settings::loadDefaults(QSettings &reg,
   m_defaultAudioTrackLanguage          = reg.value("defaultAudioTrackLanguage",    Q("und")).toString();
   m_defaultVideoTrackLanguage          = reg.value("defaultVideoTrackLanguage",    Q("und")).toString();
   m_defaultSubtitleTrackLanguage       = reg.value("defaultSubtitleTrackLanguage", Q("und")).toString();
+  m_whenToSetDefaultLanguage           = static_cast<SetDefaultLanguagePolicy>(reg.value("whenToSetDefaultLanguage",     static_cast<int>(SetDefaultLanguagePolicy::IfAbsentOrUndetermined)).toInt());
   m_defaultChapterLanguage             = reg.value("defaultChapterLanguage", Q("und")).toString();
-  m_defaultChapterCountry              = reg.value("defaultChapterCountry").toString();
+  m_defaultChapterCountry              = Util::mapToTopLevelCountryCode(reg.value("defaultChapterCountry").toString());
   auto subtitleCharset                 = reg.value("defaultSubtitleCharset").toString();
   m_defaultSubtitleCharset             = guiVersion.isEmpty() && (subtitleCharset == Q("ISO-8859-15")) ? Q("") : subtitleCharset; // Fix for a bug in versions prior to 8.2.0.
   m_defaultAdditionalMergeOptions      = reg.value("defaultAdditionalMergeOptions").toString();
   reg.endGroup();               // defaults
 }
 
+void
+Settings::loadDerivingTrackLanguagesSettings(QSettings &reg) {
+  reg.beginGroup("settings");
+  reg.beginGroup("derivingTrackLanguagesFromFileNames");
+
+  m_deriveAudioTrackLanguageFromFileNamePolicy    = static_cast<DeriveLanguageFromFileNamePolicy>(reg.value("audioPolicy",    static_cast<int>(DeriveLanguageFromFileNamePolicy::IfAbsentOrUndetermined)).toInt());
+  m_deriveVideoTrackLanguageFromFileNamePolicy    = static_cast<DeriveLanguageFromFileNamePolicy>(reg.value("videoPolicy",    static_cast<int>(DeriveLanguageFromFileNamePolicy::Never)).toInt());
+  m_deriveSubtitleTrackLanguageFromFileNamePolicy = static_cast<DeriveLanguageFromFileNamePolicy>(reg.value("subtitlePolicy", static_cast<int>(DeriveLanguageFromFileNamePolicy::IfAbsentOrUndetermined)).toInt());
+  m_regexForDerivingTrackLanguagesFromFileNames   = reg.value("customRegex").toString();
+  m_recognizedTrackLanguagesInFileNames           = reg.value("recognizedTrackLanguagesInFileNames").toStringList();
+
+  reg.endGroup();
+  reg.endGroup();
+}
+
 void
 Settings::loadSplitterSizes(QSettings &reg) {
   reg.beginGroup("splitterSizes");
@@ -264,23 +393,25 @@ Settings::loadSplitterSizes(QSettings &reg) {
   }
 
   reg.endGroup();               // splitterSizes
+}
 
-  if (m_oftenUsedLanguages.isEmpty())
-    for (auto const &languageCode : g_popular_language_codes)
-      m_oftenUsedLanguages << Q(languageCode);
+void
+Settings::loadDefaultInfoJobSettings(QSettings &reg) {
+  reg.beginGroup("settings");
+  reg.beginGroup("info");
+  reg.beginGroup("defaultJobSettings");
 
-  if (m_oftenUsedCountries.isEmpty())
-    for (auto const &countryCode : g_popular_country_codes)
-      m_oftenUsedCountries << Q(countryCode);
+  auto &s          = m_defaultInfoJobSettings;
+  s.m_mode         = static_cast<Info::JobSettings::Mode>(     reg.value("mode",      static_cast<int>(Info::JobSettings::Mode::Tree))                   .toInt());
+  s.m_verbosity    = static_cast<Info::JobSettings::Verbosity>(reg.value("verbosity", static_cast<int>(Info::JobSettings::Verbosity::StopAtFirstCluster)).toInt());
+  s.m_hexDumps     = static_cast<Info::JobSettings::HexDumps>( reg.value("hexDumps",  static_cast<int>(Info::JobSettings::HexDumps::None))               .toInt());
+  s.m_checksums    = reg.value("checksums").toBool();
+  s.m_trackInfo    = reg.value("trackInfo").toBool();
+  s.m_hexPositions = reg.value("hexPositions").toBool();
 
-  if (m_oftenUsedCharacterSets.isEmpty())
-    for (auto const &characterSet : g_popular_character_sets)
-      m_oftenUsedCharacterSets << Q(characterSet);
-
-  if (ToParentOfFirstInputFile == m_outputFileNamePolicy) {
-    m_outputFileNamePolicy = ToRelativeOfFirstInputFile;
-    m_relativeOutputDir    = Q("..");
-  }
+  reg.endGroup();
+  reg.endGroup();
+  reg.endGroup();
 }
 
 void
@@ -297,17 +428,91 @@ Settings::loadRunProgramConfigurations(QSettings &reg) {
 
     reg.beginGroup(group);
     cfg->m_active      = reg.value("active", true).toBool();
+    cfg->m_name        = reg.value("name").toString();
+    auto type          = reg.value("type", static_cast<int>(RunProgramType::ExecuteProgram)).toInt();
+    cfg->m_type        = (type > static_cast<int>(RunProgramType::Min)) && (type < static_cast<int>(RunProgramType::Max)) ? static_cast<RunProgramType>(type) : RunProgramType::Default;
+    cfg->m_forEvents   = static_cast<RunProgramForEvents>(reg.value("forEvents").toInt());
     cfg->m_commandLine = reg.value("commandLine").toStringList();
-    cfg->m_forEvents   = static_cast<RunProgramForEvents>(reg.value("forEvents").value<int>());
+    cfg->m_audioFile   = reg.value("audioFile").toString();
+    cfg->m_volume      = std::min(reg.value("volume", 50).toUInt(), 100u);
     reg.endGroup();
 
-    if (cfg->isValid())
+    if (!cfg->m_active || cfg->isValid())
       m_runProgramConfigurations << cfg;
   }
 
   reg.endGroup();               // runProgramConfigurations
 }
 
+void
+Settings::addDefaultRunProgramConfigurationForType(QSettings &reg,
+                                                   RunProgramType type,
+                                                   std::function<void(RunProgramConfig &)> const &modifier) {
+  auto guard = Q("addedDefaultConfigurationType%1").arg(static_cast<int>(type));
+
+  if (reg.value(guard).toBool() || !App::programRunner().isRunProgramTypeSupported(type))
+    return;
+
+  auto cfg      = std::make_shared<RunProgramConfig>();
+
+  cfg->m_active = true;
+  cfg->m_type   = type;
+
+  if (modifier)
+    modifier(*cfg);
+
+  if (cfg->isValid()) {
+    m_runProgramConfigurations << cfg;
+    reg.setValue(guard, true);
+  }
+}
+
+void
+Settings::addDefaultRunProgramConfigurations(QSettings &reg) {
+  reg.beginGroup("runProgramConfigurations");
+
+  auto numConfigurationsBefore = m_runProgramConfigurations.count();
+
+  addDefaultRunProgramConfigurationForType(reg, RunProgramType::PlayAudioFile, [](RunProgramConfig &cfg) { cfg.m_audioFile = App::programRunner().defaultAudioFileName(); });
+  addDefaultRunProgramConfigurationForType(reg, RunProgramType::SleepComputer);
+  addDefaultRunProgramConfigurationForType(reg, RunProgramType::HibernateComputer);
+  addDefaultRunProgramConfigurationForType(reg, RunProgramType::ShutDownComputer);
+
+  auto changed = fixDefaultAudioFileNameBug();
+
+  if ((numConfigurationsBefore != m_runProgramConfigurations.count()) || changed)
+    saveRunProgramConfigurations(reg);
+
+  reg.endGroup();               // runProgramConfigurations
+}
+
+bool
+Settings::fixDefaultAudioFileNameBug() {
+#if defined(SYS_WINDOWS)
+  // In version v11.0.0 the default audio file name is wrong:
+  // <MTX_INSTALLATION_DIRECTORY>\sounds\… instead of
+  // <MTX_INSTALLATION_DIRECTORY>\data\sounds\… where the default
+  // sound files are actually installed. As each configuration is only
+  // added once, update an existing configuration to the actual path.
+  QRegularExpression wrongFileNameRE{"<MTX_INSTALLATION_DIRECTORY>[/\\\\]sounds[/\\\\]finished-1\\.ogg"};
+  auto changed = false;
+
+  for (auto const &config : m_runProgramConfigurations) {
+    if (   (config->m_type != RunProgramType::PlayAudioFile)
+        || !config->m_audioFile.contains(wrongFileNameRE))
+      continue;
+
+    config->m_audioFile = App::programRunner().defaultAudioFileName();
+    changed             = true;
+  }
+
+  return changed;
+
+#else
+  return false;
+#endif
+}
+
 QString
 Settings::actualMkvmergeExe()
   const {
@@ -320,12 +525,17 @@ Settings::save()
   auto regPtr = registry();
   auto &reg   = *regPtr;
 
+  QVariantList enableMuxingTracksByTheseTypes;
+  for (auto type : m_enableMuxingTracksByTheseTypes)
+    enableMuxingTracksByTheseTypes << static_cast<int>(type);
+
   reg.beginGroup("info");
   reg.setValue("guiVersion",                         Q(get_current_version().to_string()));
   reg.endGroup();
 
   reg.beginGroup("settings");
   reg.setValue("priority",                           static_cast<int>(m_priority));
+  reg.setValue("probeRangePercentage",               m_probeRangePercentage);
   reg.setValue("tabPosition",                        static_cast<int>(m_tabPosition));
   reg.setValue("lastOpenDir",                        m_lastOpenDir.path());
   reg.setValue("lastOutputDir",                      m_lastOutputDir.path());
@@ -335,29 +545,39 @@ Settings::save()
   reg.setValue("oftenUsedCountries",                 m_oftenUsedCountries);
   reg.setValue("oftenUsedCharacterSets",             m_oftenUsedCharacterSets);
 
+  reg.setValue("oftenUsedLanguagesOnly",             m_oftenUsedLanguagesOnly);
+  reg.setValue("oftenUsedCountriesOnly",             m_oftenUsedCountriesOnly);
+  reg.setValue("oftenUsedCharacterSetsOnly",         m_oftenUsedCharacterSetsOnly);
+
   reg.setValue("scanForPlaylistsPolicy",             static_cast<int>(m_scanForPlaylistsPolicy));
   reg.setValue("minimumPlaylistDuration",            m_minimumPlaylistDuration);
 
   reg.setValue("setAudioDelayFromFileName",          m_setAudioDelayFromFileName);
   reg.setValue("autoSetFileTitle",                   m_autoSetFileTitle);
+  reg.setValue("autoClearFileTitle",                 m_autoClearFileTitle);
   reg.setValue("clearMergeSettings",                 static_cast<int>(m_clearMergeSettings));
   reg.setValue("disableCompressionForAllTrackTypes", m_disableCompressionForAllTrackTypes);
   reg.setValue("disableDefaultTrackForSubtitles",    m_disableDefaultTrackForSubtitles);
+  reg.setValue("mergeEnableDialogNormGainRemoval",   m_mergeEnableDialogNormGainRemoval);
   reg.setValue("mergeAlwaysShowOutputFileControls",  m_mergeAlwaysShowOutputFileControls);
   reg.setValue("mergeTrackPropertiesLayout",         static_cast<int>(m_mergeTrackPropertiesLayout));
   reg.setValue("mergeAddingAppendingFilesPolicy",    static_cast<int>(m_mergeAddingAppendingFilesPolicy));
+  reg.setValue("mergeLastAddingAppendingDecision",   static_cast<int>(m_mergeLastAddingAppendingDecision));
   reg.setValue("headerEditorDroppedFilesPolicy",     static_cast<int>(m_headerEditorDroppedFilesPolicy));
 
   reg.setValue("outputFileNamePolicy",               static_cast<int>(m_outputFileNamePolicy));
+  reg.setValue("autoDestinationOnlyForVideoFiles",   m_autoDestinationOnlyForVideoFiles);
   reg.setValue("relativeOutputDir",                  m_relativeOutputDir.path());
   reg.setValue("fixedOutputDir",                     m_fixedOutputDir.path());
   reg.setValue("uniqueOutputFileNames",              m_uniqueOutputFileNames);
+  reg.setValue("autoClearOutputFileName",            m_autoClearOutputFileName);
 
   reg.setValue("enableMuxingTracksByLanguage",       m_enableMuxingTracksByLanguage);
   reg.setValue("enableMuxingAllVideoTracks",         m_enableMuxingAllVideoTracks);
   reg.setValue("enableMuxingAllAudioTracks",         m_enableMuxingAllAudioTracks);
   reg.setValue("enableMuxingAllSubtitleTracks",      m_enableMuxingAllSubtitleTracks);
   reg.setValue("enableMuxingTracksByTheseLanguages", m_enableMuxingTracksByTheseLanguages);
+  reg.setValue("enableMuxingTracksByTheseTypes",     enableMuxingTracksByTheseTypes);
 
   reg.setValue("useDefaultJobDescription",           m_useDefaultJobDescription);
   reg.setValue("showOutputOfAllJobs",                m_showOutputOfAllJobs);
@@ -367,7 +587,6 @@ Settings::save()
   reg.setValue("removeOldJobs",                      m_removeOldJobs);
   reg.setValue("removeOldJobsDays",                  m_removeOldJobsDays);
 
-  reg.setValue("disableAnimations",                  m_disableAnimations);
   reg.setValue("showToolSelector",                   m_showToolSelector);
   reg.setValue("warnBeforeClosingModifiedTabs",      m_warnBeforeClosingModifiedTabs);
   reg.setValue("warnBeforeAbortingJobs",             m_warnBeforeAbortingJobs);
@@ -376,6 +595,7 @@ Settings::save()
 
   reg.setValue("chapterNameTemplate",                m_chapterNameTemplate);
   reg.setValue("dropLastChapterFromBlurayPlaylist",  m_dropLastChapterFromBlurayPlaylist);
+  reg.setValue("ceTextFileCharacterSet",             m_ceTextFileCharacterSet);
 
   reg.setValue("uiLocale",                           m_uiLocale);
   reg.setValue("uiFontFamily",                       m_uiFontFamily);
@@ -390,7 +610,9 @@ Settings::save()
   reg.endGroup();               // settings
 
   saveDefaults(reg);
+  saveDerivingTrackLanguagesSettings(reg);
   saveSplitterSizes(reg);
+  saveDefaultInfoJobSettings(reg);
   saveRunProgramConfigurations(reg);
 }
 
@@ -401,6 +623,7 @@ Settings::saveDefaults(QSettings &reg)
   reg.setValue("defaultAudioTrackLanguage",          m_defaultAudioTrackLanguage);
   reg.setValue("defaultVideoTrackLanguage",          m_defaultVideoTrackLanguage);
   reg.setValue("defaultSubtitleTrackLanguage",       m_defaultSubtitleTrackLanguage);
+  reg.setValue("whenToSetDefaultLanguage",           static_cast<int>(m_whenToSetDefaultLanguage));
   reg.setValue("defaultChapterLanguage",             m_defaultChapterLanguage);
   reg.setValue("defaultChapterCountry",              m_defaultChapterCountry);
   reg.setValue("defaultSubtitleCharset",             m_defaultSubtitleCharset);
@@ -408,6 +631,22 @@ Settings::saveDefaults(QSettings &reg)
   reg.endGroup();               // defaults
 }
 
+void
+Settings::saveDerivingTrackLanguagesSettings(QSettings &reg)
+  const {
+  reg.beginGroup("settings");
+  reg.beginGroup("derivingTrackLanguagesFromFileNames");
+
+  reg.setValue("audioPolicy",                         static_cast<int>(m_deriveAudioTrackLanguageFromFileNamePolicy));
+  reg.setValue("videoPolicy",                         static_cast<int>(m_deriveVideoTrackLanguageFromFileNamePolicy));
+  reg.setValue("subtitlePolicy",                      static_cast<int>(m_deriveSubtitleTrackLanguageFromFileNamePolicy));
+  reg.setValue("customRegex",                         m_regexForDerivingTrackLanguagesFromFileNames);
+  reg.setValue("recognizedTrackLanguagesInFileNames", m_recognizedTrackLanguagesInFileNames);
+
+  reg.endGroup();
+  reg.endGroup();
+}
+
 void
 Settings::saveSplitterSizes(QSettings &reg)
   const {
@@ -423,17 +662,47 @@ Settings::saveSplitterSizes(QSettings &reg)
   saveRunProgramConfigurations(reg);
 }
 
+void
+Settings::saveDefaultInfoJobSettings(QSettings &reg)
+  const {
+  reg.beginGroup("settings");
+  reg.beginGroup("info");
+  reg.beginGroup("defaultJobSettings");
+
+  auto &s = m_defaultInfoJobSettings;
+  reg.setValue("mode",         static_cast<int>(s.m_mode));
+  reg.setValue("verbosity",    static_cast<int>(s.m_verbosity));
+  reg.setValue("hexDumps",     static_cast<int>(s.m_hexDumps));
+  reg.setValue("checksums",    s.m_checksums);
+  reg.setValue("trackInfo",    s.m_trackInfo);
+  reg.setValue("hexPositions", s.m_hexPositions);
+
+  reg.endGroup();
+  reg.endGroup();
+  reg.endGroup();
+}
+
 void
 Settings::saveRunProgramConfigurations(QSettings &reg)
   const {
-  reg.remove("runProgramConfigurations");
   reg.beginGroup("runProgramConfigurations");
+
+  auto groups = reg.childGroups();
+  groups.sort();
+
+  for (auto const &group : groups)
+    reg.remove(group);
+
   auto idx = 0;
   for (auto const &cfg : m_runProgramConfigurations) {
     reg.beginGroup(Q("%1").arg(++idx, 4, 10, Q('0')));
     reg.setValue("active",      cfg->m_active);
-    reg.setValue("commandLine", cfg->m_commandLine);
+    reg.setValue("name",        cfg->m_name);
+    reg.setValue("type",        static_cast<int>(cfg->m_type));
     reg.setValue("forEvents",   static_cast<int>(cfg->m_forEvents));
+    reg.setValue("commandLine", cfg->m_commandLine);
+    reg.setValue("audioFile",   cfg->m_audioFile);
+    reg.setValue("volume",      cfg->m_volume);
     reg.endGroup();
   }
 
@@ -576,4 +845,23 @@ Settings::lastOpenDirPath()
   return Util::dirPath(m_lastOpenDir);
 }
 
+void
+Settings::runOncePerVersion(QString const &topic,
+                            std::function<void()> worker) {
+  auto reg = registry();
+  auto key = Q("runOncePerVersion/%1").arg(topic);
+
+  auto lastRunInVersion       = reg->value(key).toString();
+  auto lastRunInVersionNumber = version_number_t{to_utf8(lastRunInVersion)};
+  auto currentVersionNumber   = get_current_version();
+
+  if (   lastRunInVersionNumber.valid
+      && !(lastRunInVersionNumber < currentVersionNumber))
+    return;
+
+  reg->setValue(key, Q(currentVersionNumber.to_string()));
+
+  worker();
+}
+
 }}}
